package Gtk2::Signal;

use Inline C => Config =>
  ENABLE => AUTOWRAP,
  GLOBAL_LOAD => 1,
  NAME => 'Gtk2::Signal',
  PREFIX => 'gtkperl_signal_',
  LIBS => `pkg-config gtk+-2.0 --libs`,
  OPTIMIZE => '-g',
  INC => `pkg-config gtk+-2.0 --cflags`;

use Inline C => <<'ENDCODE';

#include <gtk/gtk.h>

/* SIGNALS */

typedef struct _perlClosure {
    GClosure closure;
    SV* callback;
    SV* extra_args; /* tuple of extra args to pass to callback */
    SV* swap_data; /* other object for gtk_signal_connect_object */
} perlClosure;

static void perl_closure_invalidate(gpointer data, GClosure *closure)
{
    perlClosure *pc = (perlClosure *)closure;
    fprintf(stderr,"Invalidating closure %lx\n", pc);
    SvREFCNT_dec(pc->callback);
    SvREFCNT_dec(pc->extra_args);
    SvREFCNT_dec(pc->swap_data);
    pc->callback = NULL;
    pc->extra_args = NULL;
    pc->swap_data = NULL;
}

static void
perl_closure_marshal(GClosure *closure,
		     GValue *return_value,
		     guint n_param_values,
		     const GValue *param_values,
		     gpointer invocation_hint,
		     gpointer marshal_data)
{
    perlClosure *pc = (perlClosure *)closure;
    guint i;
    fprintf(stderr,"Marshalling: func: %lx (refcnt: %d) data: %lx (refcnt: %d)\n", 
	    pc->callback, pc->callback->sv_refcnt,
	    pc->extra_args, pc->extra_args ? pc->extra_args->sv_refcnt : 0);
    // g_assert(pc->extra_args);
    if (pc->extra_args)
      {
	g_assert(!G_IS_OBJECT(pc->extra_args));
      }
    {
	dSP;
	ENTER;
	SAVETMPS;
	PUSHMARK(SP);
	XPUSHs(pc->extra_args ? pc->extra_args : &PL_sv_undef);
	PUTBACK ;
	perl_call_sv(pc->callback, 0);
	SPAGAIN;
	PUTBACK;
	FREETMPS;
	LEAVE;
    }
    // fprintf(stderr,"Marshalling2...\n");
}

static GClosure* perl_closure_new(SV *callback, SV *extra_args, SV *swap_data)
{
  perlClosure *closure;
  g_return_val_if_fail(callback != NULL, NULL);
  closure = (perlClosure*) g_closure_new_simple(sizeof(perlClosure), NULL);
  g_closure_add_invalidate_notifier((GClosure*) closure, NULL, perl_closure_invalidate);
  g_closure_set_marshal((GClosure*) closure, perl_closure_marshal);
  SvREFCNT_inc(callback);
  closure->callback = callback;
  closure->extra_args = NULL;
  if (extra_args && extra_args != &PL_sv_undef) {
    SvREFCNT_inc(extra_args);
    closure->extra_args = extra_args;
  }
  closure->swap_data = NULL;
  if (swap_data && swap_data != &PL_sv_undef) {
    SvREFCNT_inc(swap_data);
    closure->swap_data = swap_data;
    closure->closure.derivative_flag = TRUE;
  }
  return (GClosure*) closure;
}

long gtkperl_signal_connect(char* class, 
                            SV* objref, char* name, SV* callback, SV* data)
  {
    SV* swap_data = NULL;
    GClosure* closure = NULL;
    SvREFCNT_inc(objref);
    SvREFCNT_inc(callback);
    SvREFCNT_inc(data);
    closure = perl_closure_new(callback, data, swap_data);
    return g_signal_connect_closure(SvIV((SV*)SvRV(objref)), name, closure, FALSE);
  }



ENDCODE

1;
